import pathlib
from typing import List, Optional, Union

import aiomatrix
from . import BasePresenceStorage, engines, models


class SqlitePresenceStorage(BasePresenceStorage):
    def __init__(self, db_path: Union[pathlib.Path, str]):
        self.db_mngr = engines.SqlitePresenceStorageEngine(db_path)

    async def setup(self):
        await self.db_mngr.create_presence_table()

    async def get_user_data(
            self, account_id: str, user_id: aiomatrix.types.primitives.UserID
    ) -> Optional[models.PresenceInDB]:
        db_res = await self.db_mngr.get_user_data(account_id, user_id)
        return db_res

    async def is_new_user(self, account_id: str, user_id: aiomatrix.types.primitives.UserID) -> bool:
        event_in_db = await self.get_user_data(account_id, user_id)
        return event_in_db is None

    async def add_new_presence_update(
            self, account_id: str, event: aiomatrix.types.events.modules.presence.PresenceEvent
    ):
        if await self.is_new_user(account_id, event.sender):
            await self.db_mngr.add_new_presence(
                account_id, event.sender, event.content.presence, event.content.last_active_ago,
                event.content.status_msg
            )

    async def update_user_presence(
            self, account_id: str, user_id: aiomatrix.types.primitives.UserID,
            event: aiomatrix.types.events.modules.presence.CurrentPresenceStatus
    ):
        await self.db_mngr.update_user_presence(
            account_id, user_id, event.presence, event.last_active_ago, event.status_msg
        )

    async def get_unupdated_users(self, account_id: str, timeout: int) -> List[models.PresenceInDB]:
        db_res = await self.db_mngr.get_unupdated_users(account_id, timeout)
        return db_res



import datetime
from typing import Optional

import pydantic

from aiomatrix import types


class PresenceInDB(pydantic.BaseModel):
    user_id: types.primitives.UserID
    presence: types.misc.PresenceEnum
    last_active: Optional[datetime.datetime]
    status_msg: Optional[str]
    last_update: datetime.datetime



from typing import Union

from .room_canonical_alias import RoomCanonicalAliasContent
from .room_create import RoomCreateContent
from .room_join_rules import RoomJoinRuleEnum, RoomJoinRulesContent
from .room_member import RoomMemberContent, RoomMemberMembershipEnum
from .room_power_levels import RoomPowerLevelContent
from .room_redaction import RoomRedactionContent

RoomStateContent = Union[
    RoomCanonicalAliasContent, RoomCreateContent, RoomJoinRulesContent, RoomMemberContent, RoomPowerLevelContent, RoomRedactionContent
]



from typing import Optional

import pydantic

from aiomatrix.types import misc, primitives


class PresenceEventContent(pydantic.BaseModel):
    avatar_url: Optional[str]
    displayname: Optional[str]
    last_active_ago: Optional[int]
    presence: misc.PresenceEnum
    currently_active: Optional[bool]
    status_msg: Optional[str]


class CurrentPresenceStatus(pydantic.BaseModel):
    presence: misc.PresenceEnum
    last_active_ago: Optional[int]
    status_msg: Optional[str]
    currently_active: Optional[bool]


class PresenceEvent(pydantic.BaseModel):
    content: PresenceEventContent
    sender: primitives.UserID
    type: str



from typing import List, Optional

from aiomatrix import types
from . import raw_api


class RoomsAPI:
    def __init__(self, raw_api_client: raw_api.RawAPI):
        self._raw_api = raw_api_client

    async def create_room(
            self, room_alias: types.primitives.RoomAlias,
            visibility: Optional[types.misc.RoomVisiblityEnum] = types.misc.RoomVisiblityEnum.private,
            preset: Optional[types.misc.CreateRoomPresetEnum] = None, is_direct: bool = False,
            room_name: Optional[str] = None, topic: Optional[str] = None, invite: Optional[List[str]] = None,
            invite_3pid: Optional[List[types.methods.create_room.Invite3PID]] = None,
            room_version: Optional[str] = None, creation_content: Optional[dict] = None,
            initial_state: Optional[types.methods.create_room.InitialState] = None
    ) -> types.responses.CreateRoomResponse:
        payload = {
            'data': {
                'room_alias_name': room_alias,
                'is_direct':       is_direct
            }
        }
        if visibility is not None:
            payload['data']['visibility'] = f'{visibility}'
        if preset is not None:
            payload['data']['preset'] = f'{preset}'
        if room_name is not None:
            payload['data']['name'] = room_name
        if topic is not None:
            payload['data']['topic'] = topic
        if invite is not None:
            payload['data']['invite'] = invite
        if invite_3pid is not None:
            payload['data']['invite_3pid'] = invite_3pid
        if room_version is not None:
            payload['data']['room_version'] = room_version
        if creation_content is not None:
            payload['data']['creation_content'] = creation_content
        if initial_state is not None:
            payload['data']['initial_state'] = initial_state
        r = await self._raw_api.make_request(
            'POST', '_matrix/client/r0/createRoom', model_type=types.responses.CreateRoomResponse, **payload
        )
        return r

    async def add_new_alias(self, room_id: types.primitives.RoomID, room_alias: str):
        payload = {
            'data': {
                'room_id': room_id
            }
        }
        await self._raw_api.make_request(
            'PUT', f'_matrix/client/r0/directory/room/{room_alias}', model_type=None, **payload
        )

    async def resolve_room_alias(
            self, room_alias: types.primitives.RoomAlias
    ) -> types.responses.ResolveRoomAliasResponse:
        r = await self._raw_api.make_request(
            'GET', f'_matrix/client/r0/directory/room/{room_alias}', model_type=types.responses.ResolveRoomAliasResponse
        )
        return r

    async def delete_room_alias(self, room_alias: types.primitives.RoomAlias):
        await self._raw_api.make_request('DELETE', f'_matrix/client/r0/directory/room/{room_alias}')

    async def get_room_aliases(self, room_id: types.primitives.RoomID) -> types.responses.GetRoomAliasesResponse:
        r = await self._raw_api.make_request(
            'GET', f'_matrix/client/r0/rooms/{room_id}/aliases', model_type=types.responses.GetRoomAliasesResponse
        )
        return r



from enum import Enum

import pydantic


class RoomJoinRuleEnum(str, Enum):
    public = 'public'
    knock = 'knock'
    invite = 'invite'
    private = 'private'


class RoomJoinRulesContent(pydantic.BaseModel):
    join_rule: RoomJoinRuleEnum


# class RoomJoinRules(RoomStateEvent):
#     content: RoomJoinRulesContent



from typing import Dict, List

import pydantic


class JWK(pydantic.BaseModel):
    kty: str
    key_ops: List[str]
    alg: str
    k: str
    ext: bool


class EncryptedFile(pydantic.BaseModel):
    url: str
    key: JWK
    iv: str
    hashes: Dict[str, str]
    v: str



from __future__ import annotations

import typing

from ..filters import BaseFilter
from ..handlers import Handler, HandlerCallback
from ...client import AiomatrixClient
from ...types.events import BaseMessageEventContent, RoomEvent, RoomMessageEvent, RoomStateContent, RoomStateEvent
from ...types.events.modules.instant_messaging import NewContent


def _clear_filters(filters: typing.Tuple[BaseFilter]) -> typing.Tuple[BaseFilter]:
    """
    removing repeating filters from list
    """
    result = {}
    for i in filters:
        result[i.filter_id] = i
    return tuple(result.values())


class MatrixEventsObserver:
    def __init__(self):
        self.handlers: typing.List[Handler] = []

    def register(self, callback: typing.Callable, *filters: BaseFilter):
        _filters = _clear_filters(tuple(filters))
        self.handlers.append(
            Handler(
                callback=HandlerCallback(callback=callback),
                filters=_filters
            )
        )

    async def trigger(self, event: typing.Union[RoomEvent, RoomMessageEvent, RoomStateEvent], client: AiomatrixClient):
        additional_args: typing.Dict[
            str, typing.Optional[typing.Union[AiomatrixClient, BaseMessageEventContent, RoomStateContent]]
        ] = {}
        for handler in self.handlers:
            if await handler.check(event, client):
                for arg, annotation in handler.callback.spec.annotations.items():
                    if annotation == AiomatrixClient:
                        additional_args[arg] = client
                    elif arg == 'content':
                        additional_args['content'] = event.content
                    elif arg == 'new_content':
                        if isinstance(event, RoomMessageEvent) and isinstance(event.content, NewContent):
                            additional_args[arg] = event.content.new_content
                        else:
                            additional_args[arg] = None
                for key in list(additional_args.keys()):
                    if key not in handler.callback.spec.args:
                        del additional_args[key]
                await handler.callback.callback(event, **additional_args)



import asyncio
import json
import urllib.parse
from typing import AsyncGenerator, Dict, List, Optional, Type, TypeVar, Union

import aiohttp
import pydantic

from aiomatrix import exceptions, loggers

T = TypeVar('T')


def canonical_json_dumps(value) -> bytes:
    return json.dumps(
        value,
        # Encode code-points outside of ASCII as UTF-8 rather than \u escapes
        ensure_ascii=False,
        # Remove unnecessary white space.
        separators=(',', ':'),
        # Sort the keys of dictionaries.
        sort_keys=True,
        # Encode the resulting Unicode as UTF-8 bytes.
    ).encode("UTF-8")


class RawAPI:
    def __init__(self, server_url: str):
        self._session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=60))
        self._BASE_URL = server_url
        self._access_token = ''

    async def __make_request(self, http_method: str, method: str, **kwargs) -> Optional[Dict]:
        args = {
            'method':  http_method,
            'url':     f'{self._BASE_URL}/{urllib.parse.quote(method)}',
            'headers': {
                'Authorization': f'Bearer {self._access_token}',
                'Content-Type':  'application/json'
            },
            **kwargs
        }
        if 'data' in args and isinstance(args['data'], dict):
            args['data'] = canonical_json_dumps(args['data'])
        else:
            del args['headers']['Content-Type']
        if self._access_token == '':
            del args['headers']['Authorization']
        try:
            async with self._session.request(**args) as resp:
                res = await resp.text()
                json_response = json.loads(res)
                if not resp.ok:
                    if 'errcode' in res:
                        raise exceptions.MatrixAPIError.detect(
                            json_response['errcode'], json_response['error'], json_response
                        )
        except asyncio.TimeoutError:
            raise exceptions.MatrixAPINetworkError('Request timeout error')
        except aiohttp.ClientError as e:
            raise exceptions.MatrixAPINetworkError(f"{type(e).__name__}: {e}")
        else:
            return json_response

    @staticmethod
    def _convert_to_model(data: Optional[dict], model: Type[T]) -> Optional[T]:
        try:
            return model(**data)
        except pydantic.ValidationError as e:
            loggers.client.warn(f'validation error {e=} for json {data=}')
            return None

    async def make_request(
            self, http_method: str, method: str, model_type: Optional[Type[T]] = None, **kwargs
    ) -> Optional[Union[List[T], T]]:
        r: Optional[dict] = await self.__make_request(http_method, method, **kwargs)
        if r is not None:
            if model_type is not None:
                if isinstance(r, dict):
                    return self._convert_to_model(r, model_type)
                elif isinstance(r, list) and isinstance(r[0], dict):
                    return [self._convert_to_model(i, model_type) for i in r]
            else:
                return r
        else:
            return None

    async def stream_response(self, method: str, chunk_size: int, **kwargs) -> AsyncGenerator[bytes, None]:
        args = {
            'method':  'GET',
            'url':     f'{self._BASE_URL}/{urllib.parse.quote(method)}',
            'headers': {
                'Authorization': f'Bearer {self._access_token}',
            },
            **kwargs
        }
        if 'data' in args:
            args['data'] = canonical_json_dumps(args['data'])
        if self._access_token == '':
            del args['headers']['Authorization']
        try:
            async with self._session.request(**args) as resp:
                if resp.status == 200:
                    async for chunk in resp.content.iter_chunked(chunk_size):
                        yield chunk
                else:
                    res = await resp.text()
                    json_response = json.loads(res)
                    raise exceptions.MatrixAPIError.detect(
                        json_response['errcode'], json_response['error'], json_response
                    )
        except asyncio.TimeoutError:
            raise exceptions.MatrixAPINetworkError('Request timeout error')
        except aiohttp.ClientError as e:
            raise exceptions.MatrixAPINetworkError(f"{type(e).__name__}: {e}")

    @property
    def access_token(self):
        return self._access_token

    @access_token.setter
    def access_token(self, token: str):
        self._access_token = token

    @property
    def is_authorized(self) -> bool:
        return self.access_token != ''



from __future__ import annotations

from .audio import AudioContent
from .emote import EmoteContent
from .file import FileContent
from .image import ImageContent
from .location import LocationContent
from .new_content import NewContent
from .notice import NoticeContent
from .text import TextContent
from .video import VideoContent

__all__ = [
    'TextContent',
    'NoticeContent',
    'EmoteContent',
    'AudioContent',
    'FileContent',
    'ImageContent',
    'LocationContent',
    'VideoContent',
    'NewContent'
]

for _entity_name in __all__:
    _entity = globals()[_entity_name]
    if not hasattr(_entity, "update_forward_refs"):
        continue
    _entity.update_forward_refs(**globals())

del _entity
del _entity_name



from __future__ import annotations

from typing import Optional

from .basic_file_info import BasicFileInfo
from ..e2ee import EncryptedFile
from ...base_room_events import BaseMessageEventContent


class AudioInfo(BasicFileInfo):
    duration: int


class AudioContent(BaseMessageEventContent):
    msgtype: str = 'm.audio'
    info: Optional[AudioInfo]
    url: Optional[str]
    file: Optional[EncryptedFile]



from enum import Enum


class RoomVisiblityEnum(str, Enum):
    public = 'public'
    private = 'private'


class CreateRoomPresetEnum(str, Enum):
    public_chat = 'public_chat'
    private_chat = 'private_chat'
    trusted_private_chat = 'trusted_private_chat'


class RoomStabilityEnum(str, Enum):
    stable = 'stable'
    unstable = 'unstable'


class RoomEventTypesEnum(str, Enum):
    room_message = 'm.room.message'
    room_member = 'm.room.member'
    reaction = 'm.reaction'
    redaction = 'm.room.redaction'


class RoomMessageEventMsgTypesEnum(str, Enum):
    text = 'm.text'
    emote = 'm.emote'  # looks unused, replaced with m.reaction
    reaction = 'm.reaction'
    notice = 'm.notice'
    image = 'm.image'
    file = 'm.file'
    audio = 'm.audio'
    location = 'm.location'
    video = 'm.video'


class PresenceEnum(str, Enum):
    offline = 'offline'
    online = 'online'
    unavailable = 'unavailable'


class ResizingMethodsEnum(str, Enum):
    crop = 'crop'
    scale = 'scale'



from .internal_data import BaseInternalDataStorage, SqliteInternalDataStorage
from .presence import BasePresenceStorage, SqlitePresenceStorage
from .room_events import BaseEventStorage, SqliteEventStorage
from .storage import StorageRepo



from typing import Optional

from pydantic import BaseModel, Field


class ContentRepositoryConfig(BaseModel):
    upload_size: Optional[int] = Field(None, alias='m.upload.size')


class UploadedFileResponse(BaseModel):
    content_uri: str



from __future__ import annotations

from typing import Optional

from ...base_room_events import BaseMessageEventContent


class TextContent(BaseMessageEventContent):
    format: Optional[str]
    formatted_body: Optional[str]



from typing import Dict, List, Optional, Union

import aiomatrix
from . import BaseEventStorage, models


class EventsRepo:
    def __init__(self, storage: BaseEventStorage):
        self.storage = storage

    async def setup(self):
        await self.storage.setup()

    async def get_event_data(
            self, account_id: Union[str, aiomatrix.types.responses.WhoAmIResponse],
            event_id: aiomatrix.types.primitives.EventID
    ) -> Optional[models.EventInDB]:
        if isinstance(account_id, aiomatrix.types.responses.WhoAmIResponse):
            account_id = account_id.user_id
        is_seen_event = event_id in (await self.storage.check_seen_events(account_id, [event_id]))
        if is_seen_event:
            return await self.storage.get_event_data(event_id)
        else:
            return None

    async def get_event_data_batch(
            self, account_id: Union[str, aiomatrix.types.responses.WhoAmIResponse],
            events_ids: List[aiomatrix.types.primitives.EventID]
    ) -> Dict[str, models.EventInDB]:
        if isinstance(account_id, aiomatrix.types.responses.WhoAmIResponse):
            account_id = account_id.user_id
        seen_events = list(map(lambda x: x.event_id, await self.storage.check_seen_events(account_id, events_ids)))
        return await self.storage.get_multiple_events_data(seen_events)

    async def is_new_event(
            self, account_id: Union[str, aiomatrix.types.responses.WhoAmIResponse],
            event_id: aiomatrix.types.primitives.EventID
    ) -> bool:
        if isinstance(account_id, aiomatrix.types.responses.WhoAmIResponse):
            account_id = account_id.user_id
        is_seen_event = event_id in (await self.storage.check_seen_events(account_id, [event_id]))
        if is_seen_event:
            return await self.storage.is_new_event(event_id)
        else:
            return True

    async def are_new_events(
            self, account_id: Union[str, aiomatrix.types.responses.WhoAmIResponse],
            events_ids: List[aiomatrix.types.primitives.EventID]
    ) -> Dict[str, bool]:
        if isinstance(account_id, aiomatrix.types.responses.WhoAmIResponse):
            account_id = account_id.user_id
        events_data = await self.storage.check_seen_events(account_id, events_ids)
        found_ids = set(map(lambda x: x.event_id, events_data))
        res = {i: i in found_ids for i in events_ids}
        return res

    async def insert_new_event(
            self, account_id: Union[str, aiomatrix.types.responses.WhoAmIResponse],
            event: aiomatrix.types.events.RoomEvent
    ):
        if isinstance(account_id, aiomatrix.types.responses.WhoAmIResponse):
            account_id = account_id.user_id
        await self.storage.add_new_seen_event(account_id, event.event_id)
        return await self.storage.insert_new_event(event)

    async def insert_new_events_batch(
            self, account_id: Union[str, aiomatrix.types.responses.WhoAmIResponse],
            events: List[aiomatrix.types.events.RoomEvent]
    ):
        if isinstance(account_id, aiomatrix.types.responses.WhoAmIResponse):
            account_id = account_id.user_id
        await self.storage.add_new_seen_events_batch(account_id, list(map(lambda x: x.event_id, events)))
        return await self.storage.insert_new_events_batch(events)

    # TODO: implement with account_id in mind
    # async def get_events_by_room(
    #         self, account_id: Union[str, aiomatrix.test_types.responses.WhoAmIResponse],
    #         room_id: aiomatrix.test_types.primitives.RoomID
    # ) -> List[models.EventInDB]:
    #     if isinstance(account_id, aiomatrix.test_types.responses.WhoAmIResponse):
    #         account_id = account_id.user_id
    #     return await self.storage.get_events_by_room(account_id, room_id)



from typing import List

import pydantic

from .. import primitives


class UserJoinedRoomsResponse(pydantic.BaseModel):
    joined_rooms: List[primitives.RoomID]


class UserJoinRoomResponse(pydantic.BaseModel):
    room_id: primitives.RoomID



import typing
from typing import Optional, Tuple

from . import apis
from .apis import raw_api
from .. import types
from ..utils.mixins import ContextVarMixin


class AiomatrixClient(ContextVarMixin):
    def __init__(self, server_url: str, auth_details: Tuple[str, dict]):
        self.server_url = server_url
        self._auth_cb = {'password': self.login_by_password}[auth_details[0]]
        self._auth_details = auth_details[1]
        self._raw_api = raw_api.RawAPI(self.server_url)
        self.auth_api = apis.AuthAPI(self._raw_api)
        self.capabilities_api = apis.CapabilitiesAPI(self._raw_api)
        self.room_api = apis.RoomsAPI(self._raw_api)
        self.room_membership_api = apis.RoomMembershipAPI(self._raw_api)
        self.listing_room_api = apis.ListingRoomsAPI(self._raw_api)
        self.sync_api = apis.SyncingAPI(self._raw_api)

        self.presence_api = apis.modules.PresenceAPI(self._raw_api)
        self.instant_messaging_api = apis.modules.InstantMessagingAPI(self.sync_api)
        self.typing_notifications_api = apis.modules.TypingNotifications(self._raw_api)
        self.content_repository_api = apis.modules.ContentRepositoryAPI(self._raw_api)

        self.me: typing.Optional[types.responses.WhoAmIResponse] = None

    async def login(self):
        if not self._raw_api.is_authorized:
            login_result = await self._auth_cb(**self._auth_details)
            if login_result:
                self.me = await self.auth_api.whoami()

    async def login_by_password(self, login: str, password: str, device_id: Optional[str] = None) -> bool:
        supported_login_types = await self.auth_api.get_login_types()
        support_password_auth = 'm.login.password' in map(lambda x: x.type, supported_login_types.flows)
        if support_password_auth:
            login_response = await self.auth_api.password_login(login, password, device_id)
            if login_response.access_token:
                self.auth_api.set_access_token(login_response.access_token)
                return True
        return False



from typing import Optional

from aiomatrix import types
from aiomatrix.utils import quotes
from .. import raw_api


class PresenceAPI:
    def __init__(self, raw_api_client: raw_api.RawAPI):
        self._raw_api = raw_api_client

    async def set_user_state(
            self, user_id: types.primitives.UserID, new_presence: types.misc.PresenceEnum,
            status_msg: Optional[str] = None
    ):
        payload = {
            'data': {
                'presence': new_presence
            }
        }
        if status_msg is not None:
            payload['data']['status_msg'] = status_msg
        await self._raw_api.make_request(
            'PUT', f'_matrix/client/r0/presence/{quotes.quote_user_id(user_id)}/status',
            model_type=types.events.modules.presence.CurrentPresenceStatus
        )

    async def get_user_presence(
            self, user_id: types.primitives.UserID
    ) -> types.events.modules.presence.CurrentPresenceStatus:
        r = await self._raw_api.make_request(
            'GET',
            f'_matrix/client/r0/presence/{quotes.quote_user_id(user_id)}/status',
            model_type=types.events.modules.presence.CurrentPresenceStatus
        )
        return r



from . import modules, relationships
from .base_room_events import (
    BasicEvent, BaseMessageEventContent, RelationshipMixin, RelationshipToEventData, RoomEvent, RoomStateEvent
)
from .room_message_event import RoomMessageEvent
from .room_state_events import (
    RoomCanonicalAliasContent, RoomCreateContent, RoomJoinRuleEnum, RoomJoinRulesContent, RoomMemberContent,
    RoomMemberMembershipEnum, RoomPowerLevelContent, RoomRedactionContent, RoomStateContent
)



import datetime
from typing import List, Optional

import aiomatrix
from .. import models
from ...base_engines import SqliteConnection


class SqlitePresenceStorageEngine(SqliteConnection):
    async def create_presence_table(self):
        sql = '''
            create table if not exists presence
            (
                account_id  TEXT,
                user_id     TEXT,
                presence    TEXT,
                last_active INTEGER,
                status_msg  TEXT,
                last_update INTEGER,
                UNIQUE (account_id, user_id) ON CONFLICT IGNORE,
                PRIMARY KEY (account_id, user_id)
            );
        '''
        self._make_request(sql)

    async def get_user_data(
            self, account_id: str, user_id: aiomatrix.types.primitives.UserID
    ) -> Optional[models.PresenceInDB]:
        sql = 'SELECT * FROM presence WHERE user_id = ? AND account_id = ?'
        params = (user_id, account_id)
        r = self._make_request(sql, params, fetch=True, model_type=models.PresenceInDB)
        return r

    async def add_new_presence(
            self, account_id: str, user_id: aiomatrix.types.primitives.UserID,
            presence: aiomatrix.types.misc.PresenceEnum, last_active_ago: Optional[int] = None,
            status_msg: Optional[str] = None
    ):
        now = datetime.datetime.utcnow()
        sql = 'INSERT INTO presence (user_id, presence, last_active, status_msg, last_update, account_id) VALUES (?, ?, ?, ?, ?, ?)'
        params = (
            user_id, f'{presence}',
            datetime.datetime.fromtimestamp(now.timestamp() - last_active_ago).isoformat() if last_active_ago else None,
            status_msg, now.isoformat(), account_id
        )
        self._make_request(sql, params)

    async def update_user_presence(
            self, account_id: str, user_id: aiomatrix.types.primitives.UserID,
            presence: aiomatrix.types.misc.PresenceEnum,
            last_active_ago: Optional[int] = None, status_msg: Optional[str] = None
    ):
        now = datetime.datetime.utcnow()
        sql = 'UPDATE presence SET presence = ?, last_active = ?, status_msg = ?, last_update = ? WHERE user_id = ? AND account_id = ?'
        params = (
            f'{presence}',
            datetime.datetime.fromtimestamp(now.timestamp() - last_active_ago).isoformat() if last_active_ago else None,
            status_msg, now.isoformat(), user_id, account_id
        )
        self._make_request(sql, params)

    async def get_unupdated_users(self, account_id: str, timeout: int) -> List[models.PresenceInDB]:
        sql = 'SELECT * FROM presence WHERE last_active + ? < ? AND account_id = ?'
        params = (timeout, datetime.datetime.utcnow().timestamp() + timeout, account_id)
        r = self._make_request(sql, params, fetch=True, mult=True, model_type=models.PresenceInDB)
        return r



from aiomatrix import types
from . import raw_api


class CapabilitiesAPI:
    def __init__(self, raw_api_client: raw_api.RawAPI):
        self._raw_api = raw_api_client

    async def get_server_capabilities(self) -> types.responses.ServerCapabilitiesResponse:
        r = await self._raw_api.make_request(
            'GET', '_matrix/client/r0/capabilities', model_type=types.responses.ServerCapabilitiesResponse
        )
        return r



import math
from typing import List, Optional


class AiomatrixError(Exception):
    description_url: Optional[str] = None

    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        message = self.message
        if self.description_url:
            message += f"\n(more info: {self.description_url})"
        return message


class MatrixAPIError(AiomatrixError):
    __suberrors: List['MatrixAPIError'] = []
    error_code = None
    error = None
    raw_data = None

    def __init__(self, error_code: int, error_text: str, raw_data: dict):
        super().__init__(f'{error_code} - {error_text}')
        self.error_code = error_code
        self.error = error_text
        self.raw_data = raw_data

    def __init_subclass__(cls, error_code=None, **kwargs):
        super(MatrixAPIError, cls).__init_subclass__(**kwargs)
        if error_code is not None:
            cls.error_code = error_code.upper()
            # noinspection PyTypeChecker
            cls.__suberrors.append(cls)

    @classmethod
    def detect(cls, error_code: str, error: str, raw_data: dict):
        match = error_code.upper()
        for err in cls.__suberrors:
            if err is cls:
                continue
            if err.error_code == match:
                # noinspection PyCallingNonCallable
                raise err(error_code, error, raw_data)
        raise cls(error_code, error, raw_data)


class MatrixAPINetworkError(MatrixAPIError):
    pass


class MissingToken(MatrixAPIError, error_code='M_MISSING_TOKEN'):
    pass


class UnknownToken(MatrixAPIError, error_code='M_UNKNOWN_TOKEN'):
    pass


class Forbidden(MatrixAPIError, error_code='M_FORBIDDEN'):
    pass


class RoomAliasInUse(MatrixAPIError, error_code='M_ROOM_IN_USE'):
    pass


class Unrecognized(MatrixAPIError, error_code='M_UNRECOGNIZED'):
    pass


class RateLimit(MatrixAPIError, error_code='M_LIMIT_EXCEEDED'):
    description_url = 'https://spec.matrix.org/v1.1/client-server-api/#common-error-codes'

    def __init__(self, error_code: int, error_text: str, raw_data: dict, ):
        super().__init__(error_code, error_text, raw_data)
        self.retry_after_ms = raw_data['retry_after_ms']
        self.retry_after = math.ceil(self.retry_after_ms // 1000)



import pathlib
from typing import Dict, List, Optional, Union

import aiomatrix
from . import BaseEventStorage, engines, models


class SqliteEventStorage(BaseEventStorage):
    def __init__(self, db_path: Union[pathlib.Path, str]):
        self.db_mngr = engines.SqliteEventStorageEngine(db_path)

    async def setup(self):
        await self.db_mngr.create_seen_events_table()
        await self.db_mngr.create_events_table()

    async def get_event_data(
            self, event_id: aiomatrix.types.primitives.EventID
    ) -> Optional[models.EventInDB]:
        db_res = await self.db_mngr.get_event(event_id)
        return db_res

    async def is_new_event(self, event: aiomatrix.types.events.RoomEvent) -> bool:
        event_in_db = await self.get_event_data(event.event_id)
        return event_in_db is None

    async def insert_new_event(self, event: aiomatrix.types.events.RoomEvent):
        await self.db_mngr.add_new_state_event(event)

    async def insert_new_events_batch(self, events: List[aiomatrix.types.events.RoomEvent]):
        await self.db_mngr.add_new_state_events_batch(events)

    async def get_events_by_room(
            self, room_id: aiomatrix.types.primitives.RoomID
    ) -> List[models.EventInDB]:
        db_res = await self.db_mngr.get_room_events(room_id)
        return db_res

    async def get_multiple_events_data(
            self, events_ids: List[aiomatrix.types.primitives.EventID]
    ) -> Dict[str, models.EventInDB]:
        db_res = await self.db_mngr.get_events_batch(events_ids)
        res = {i.event_id: i for i in db_res}
        return res

    async def get_account_seen_events(self, account_id: str) -> List[models.SeenEvent]:
        db_res = await self.db_mngr.get_account_seen_events(account_id)
        return db_res

    async def check_seen_events(
            self, account_id: str, event_ids: List[aiomatrix.types.primitives.EventID]
    ) -> List[models.SeenEvent]:
        db_res = await self.db_mngr.check_seen_events(account_id, event_ids)
        return db_res

    async def add_new_seen_event(self, account_id: str, event_id: aiomatrix.types.primitives.EventID):
        await self.db_mngr.add_new_seen_event(account_id, event_id)

    async def add_new_seen_events_batch(self, account_id: str, event_ids: List[aiomatrix.types.primitives.EventID]):
        await self.db_mngr.add_new_seen_events_batch(account_id, event_ids)



from typing import Any

import pydantic


class InitialState(pydantic.BaseModel):
    type: str
    state_key: str = ''
    content: Any


class Invite3PID(pydantic.BaseModel):
    id_server: str
    id_access_token: str
    medium: str
    address: str



from typing import Optional

from . import models


class BaseInternalDataStorage:
    async def setup(self):
        raise NotImplementedError

    async def get_last_sync_token(self, account_id: str) -> Optional[models.InternalDataPair]:
        raise NotImplementedError

    async def set_last_sync_token(self, account_id: str, token: str):
        raise NotImplementedError



import asyncio
from asyncio import AbstractEventLoop
from typing import Callable, List, Union

import aiomatrix


class Executor:
    def __init__(self, client: 'aiomatrix.AiomatrixDispatcher', loop: AbstractEventLoop = None):
        if loop is not None:
            self._loop = loop
        self.client = client
        self._on_startup = []
        self._on_shutdown = []
        self._running_lock = asyncio.Lock()

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return getattr(self, '_loop', asyncio.get_event_loop())

    def on_startup(self, callback: Union[Callable, List[Callable]]):
        if not isinstance(callback, Callable):
            return
        if isinstance(callback, list):
            for cb in callback:
                self.on_startup(cb)
            return
        self._on_startup.append(callback)

    def on_shutdown(self, callback: Union[Callable, List[Callable]]):
        if not isinstance(callback, Callable):
            return
        if isinstance(callback, list):
            for cb in callback:
                self.on_shutdown(cb)
            return
        self._on_shutdown.append(callback)

    def start_polling(
            self, ignore_errors: bool = True, track_presence: bool = False, timeout: int = 10, sleep: float = 0.1,
    ):
        loop: AbstractEventLoop = self.loop

        try:
            loop.run_until_complete(self._start_polling())
            loop.create_task(self.client.run_polling(ignore_errors, timeout, sleep, track_presence))
            loop.run_forever()
        except (KeyboardInterrupt, SystemExit):
            # loop.stop()
            pass
        finally:
            loop.run_until_complete(self._shutdown_polling())

    async def _start_polling(self):
        for cb in self._on_startup:
            await cb(self.client)

    async def _shutdown_polling(self):
        for cb in self._on_shutdown:
            await cb(self.client)



from aiomatrix import types
from aiomatrix.utils import quotes
from .. import raw_api


class TypingNotifications:
    def __init__(self, raw_api_client: raw_api.RawAPI):
        self._raw_api = raw_api_client

    async def send_typing(
            self, user_id: types.primitives.UserID, room_id: types.primitives.RoomID, timeout: int
    ):
        payload = {
            'data': {
                'timeout': timeout
            }
        }
        await self._raw_api.make_request(
            'PUT', f'_matrix/client/r0/rooms/{room_id}/typing/{quotes.quote_user_id(user_id)}',
            **payload
        )

    async def send_typing_status(
            self, user_id: types.primitives.UserID, room_id: types.primitives.RoomID, status: bool
    ):
        payload = {
            'data': {
                'status': status
            }
        }
        await self._raw_api.make_request(
            'PUT', f'_matrix/client/r0/rooms/{room_id}/typing/{quotes.quote_user_id(user_id)}',
            **payload
        )



from typing import List, Optional, Union

from aiomatrix import types
from . import raw_api


class RoomMembershipAPI:
    def __init__(self, raw_api_client: raw_api.RawAPI):
        self._raw_api = raw_api_client

    async def get_user_joined_rooms(self) -> types.responses.UserJoinedRoomsResponse:
        r = await self._raw_api.make_request(
            'GET', f'_matrix/client/r0/joined_rooms', model_type=types.responses.UserJoinedRoomsResponse
        )
        return r

    async def invite_to_room(self, room_id: types.primitives.RoomID, user_id: types.primitives.UserID):
        payload = {
            'data': {
                'user_id': user_id
            }
        }
        await self._raw_api.make_request(
            'POST', f'_matrix/client/r0/rooms/{room_id}/invite', model_type=None, **payload
        )

    async def join_room_by_id(
            self, room_id: types.primitives.RoomID,
            third_party_signed: Optional[types.methods.join_room_query.ThirdPartySigned] = None
    ) -> types.responses.UserJoinRoomResponse:
        payload = {'data': {}}
        if third_party_signed is not None:
            payload['data']['third_party_signed'] = third_party_signed.dict()
        r = await self._raw_api.make_request(
            'POST', f'_matrix/client/r0/rooms/{room_id}/join', model_type=types.responses.UserJoinRoomResponse,
            **payload
        )
        return r

    async def join_room(
            self, room_id_or_alias: Union[types.primitives.RoomID, types.primitives.RoomAlias],
            servers: Optional[List[str]] = None,
            third_party_signed: Optional[types.methods.join_room_query.ThirdPartySigned] = None
    ) -> types.responses.UserJoinRoomResponse:
        payload = {'data': {}, 'params': {}}
        if third_party_signed is not None:
            payload['data']['third_party_signed'] = third_party_signed.dict()
        if servers is not None:
            payload['params'] = [('server_name', i) for i in servers]
        r = await self._raw_api.make_request(
            'POST', f'_matrix/client/r0/join/{room_id_or_alias}', model_type=types.responses.UserJoinRoomResponse,
            **payload
        )
        return r

    async def leave_room(self, room_id: types.primitives.RoomID):
        await self._raw_api.make_request('POST', f'_matrix/client/r0/rooms/{room_id}/leave')

    async def forget_room(self, room_id: types.primitives.RoomID):
        await self._raw_api.make_request('POST', f'_matrix/client/r0/rooms/{room_id}/forget')

    async def kick_from_room(
            self, room_id: types.primitives.RoomID, user_id: types.primitives.UserID, reason: Optional[str] = None
    ):
        payload = {
            'data': {
                'user_id': user_id
            }
        }
        if reason is not None:
            payload['data']['reason'] = reason
        await self._raw_api.make_request(
            'POST', f'_matrix/client/r0/rooms/{room_id}/kick', model_type=None, **payload
        )

    async def ban_from_room(
            self, room_id: types.primitives.RoomID, user_id: types.primitives.UserID, reason: Optional[str] = None
    ):
        payload = {
            'data': {
                'user_id': user_id
            }
        }
        if reason is not None:
            payload['data']['reason'] = reason
        await self._raw_api.make_request(
            'POST', f'_matrix/client/r0/rooms/{room_id}/ban', model_type=None, **payload
        )

    async def unban_in_room(self, room_id: types.primitives.RoomID, user_id: types.primitives.UserID):
        payload = {
            'data': {
                'user_id': user_id
            }
        }
        await self._raw_api.make_request(
            'POST', f'_matrix/client/r0/rooms/{room_id}/unban', model_type=None, **payload
        )



from typing import NewType

RoomID = NewType('RoomID', str)
RoomAlias = NewType('RoomAlias', str)

UserID = NewType('UserID', str)
EventID = NewType('EventID', str)

# class ContentMixin:
#     content:
EventContent = NewType('EventContent', dict)



from pydantic import Field

from ..base_room_events import RelationshipMixin, RelationshipToEventData


class Reaction(RelationshipToEventData):
    key: str


class ReactionRelationshipContent(RelationshipMixin):
    relationship: Reaction = Field(..., alias='m.relates_to')



from __future__ import annotations

import typing

from pydantic import Field, validator

from aiomatrix.types.misc import RoomMessageEventMsgTypesEnum
from ...base_room_events import BaseMessageEventContent

if typing.TYPE_CHECKING:
    from .audio import AudioContent
    from .emote import EmoteContent
    from .file import FileContent
    from .image import ImageContent
    from .location import LocationContent
    from .notice import NoticeContent
    from .video import VideoContent
    from .text import TextContent


class NewContent(BaseMessageEventContent):
    new_content: typing.Union[
        AudioContent, EmoteContent, FileContent, ImageContent, LocationContent, NoticeContent, TextContent, VideoContent
    ] = Field(..., alias='m.new_content')

    @validator('new_content')
    def _fix_typing(cls, v: typing.Union[
        AudioContent, EmoteContent, FileContent, ImageContent, LocationContent, NoticeContent, TextContent, VideoContent
    ]):
        from .audio import AudioContent
        from .emote import EmoteContent
        from .file import FileContent
        from .image import ImageContent
        from .location import LocationContent
        from .notice import NoticeContent
        from .video import VideoContent
        from .text import TextContent
        supported_msgtypes = {
            RoomMessageEventMsgTypesEnum.audio:    AudioContent,
            RoomMessageEventMsgTypesEnum.emote:    EmoteContent,
            RoomMessageEventMsgTypesEnum.file:     FileContent,
            RoomMessageEventMsgTypesEnum.image:    ImageContent,
            RoomMessageEventMsgTypesEnum.location: LocationContent,
            RoomMessageEventMsgTypesEnum.notice:   NoticeContent,
            RoomMessageEventMsgTypesEnum.text:     TextContent,
            RoomMessageEventMsgTypesEnum.video:    VideoContent,
        }
        content = supported_msgtypes[v.msgtype](**v.raw)
        return content



from enum import Enum
from typing import Optional

import pydantic


class RoomMemberMembershipEnum(str, Enum):
    invite = 'invite'
    join = 'join'
    knock = 'knock'
    leave = 'leave'
    ban = 'ban'


class RoomMemberInvite(pydantic.BaseModel):
    display_name: str
    signed: dict


class RoomMemberUnsignedData(pydantic.BaseModel):
    invite_room_state: dict


class RoomMemberContent(pydantic.BaseModel):
    avatar_url: Optional[str]
    displayname: Optional[str]
    membership: RoomMemberMembershipEnum
    is_direct: bool
    third_party_invite: RoomMemberInvite
    unsigned: RoomMemberUnsignedData


# class RoomMemberEvent(RoomStateEvent):
#     content: RoomMemberContent



import pydantic
from pydantic import Field

from ..primitives import UserID


class HomeserverInformation(pydantic.BaseModel):
    base_url: str


class IdentityInformation(pydantic.BaseModel):
    base_url: str


class DiscoveryInformation(pydantic.BaseModel):
    homeserver: HomeserverInformation = Field(..., alias='m.homeserver')
    identity_server: IdentityInformation = Field(None, alias='m.identity_server')


class LoginResponse(pydantic.BaseModel):
    user_id: UserID
    access_token: str
    device_id: str
    well_known: DiscoveryInformation



from typing import List, Optional

from .. import models
from ...base_engines import SqliteConnection


class SqliteInternalDataStorageEngine(SqliteConnection):
    async def create_internal_data_table(self):
        sql = '''
            create table if not exists internal_data
            (
                account_id TEXT,
                key        TEXT,
                data       TEXT,
                UNIQUE (account_id, key) ON CONFLICT ABORT,
                PRIMARY KEY (account_id, key)
            );
        '''
        self._make_request(sql)

    async def get_data_by_key(self, account_id: str, key: str) -> Optional[models.InternalDataPair]:
        sql = 'SELECT * FROM internal_data WHERE key = ? AND account_id = ?'
        params = (key, account_id)
        return self._make_request(sql, params, fetch=True, model_type=models.InternalDataPair)

    async def get_all_data(self, account_id: str) -> List[models.InternalDataPair]:
        sql = 'SELECT * FROM internal_data WHERE account_id = ?'
        params = (account_id,)
        return self._make_request(sql, params=params, fetch=True, mult=True, model_type=models.InternalDataPair)

    async def insert_data(self, account_id: str, key: str, value: str):
        sql = 'INSERT INTO internal_data (key, data, account_id) VALUES (?, ?, ?)'
        params = (key, f'{value}', account_id)
        self._make_request(sql, params)

    async def set_data_by_key(self, account_id: str, key: str, value: str):
        sql = 'UPDATE internal_data SET data = ? WHERE key = ?AND account_id = ?'
        params = (f'{value}', key, account_id)
        self._make_request(sql, params)

    async def delete_data_by_key(self, account_id: str, key: str):
        sql = 'DELETE FROM internal_data WHERE key = ? AND account_id = ?'
        params = (key, account_id)
        self._make_request(sql, params)



from typing import Dict, List, Optional

import pydantic
from pydantic import Field

from ..events import BasicEvent, RoomEvent, RoomStateEvent, modules
from ..primitives import EventID, RoomID, UserID


class RoomSummary(pydantic.BaseModel):
    heroes: Optional[List[str]] = Field(None, alias='m.heroes')
    joined_member_count: Optional[int] = Field(None, alias='m.joined_member_count')
    invited_member_count: Optional[int] = Field(None, alias='m.invited_member_count')


class UnreadNotifications(pydantic.BaseModel):
    highlight_count: int
    notification_count: int


class Timeline(pydantic.BaseModel):
    events: List[RoomEvent]
    limited: bool
    prev_batch: str


class State(pydantic.BaseModel):
    events: List[RoomStateEvent]


class AccountData(pydantic.BaseModel):
    events: List[BasicEvent]


class Presence(pydantic.BaseModel):
    events: List[modules.presence.PresenceEvent]


class Ephemeral(pydantic.BaseModel):
    events: List[BasicEvent]


class JoinedRooms(pydantic.BaseModel):
    summary: Optional[RoomSummary]
    state: Optional[State]
    timeline: Optional[Timeline]
    ephemeral: Optional[Ephemeral]
    account_data: Optional[AccountData]
    unread_notifications: Optional[UnreadNotifications]


class Event(BasicEvent):
    sender: UserID


class StrippedStateEvent(Event):
    state_key: str


class InviteState(pydantic.BaseModel):
    events: List[StrippedStateEvent]


class InvitedRooms(pydantic.BaseModel):
    invite_state: InviteState


class LeftRooms(pydantic.BaseModel):
    state: State
    timeline: Timeline
    account_data: AccountData


class Rooms(pydantic.BaseModel):
    join: Optional[Dict[RoomID, JoinedRooms]]
    invite: Optional[Dict[RoomID, InvitedRooms]]
    leave: Optional[Dict[RoomID, LeftRooms]]


class SyncResponse(pydantic.BaseModel):
    next_batch: str
    rooms: Optional[Rooms]
    presence: Optional[Presence]
    account_data: Optional[AccountData]
    to_device: Optional[modules.send_to_device.ToDevice]
    device_lists: Optional[modules.e2ee.DeviceLists]
    device_one_time_keys_count: Optional[Dict[str, int]]


class SentEventResponse(pydantic.BaseModel):
    event_id: EventID



import inspect
import typing

import pydantic
from pydantic import root_validator

from aiomatrix import types
from aiomatrix.dispatcher.filters import BaseFilter

if typing.TYPE_CHECKING:
    from aiomatrix import AiomatrixClient


class HandlerCallback(pydantic.BaseModel):
    callback: typing.Callable
    spec: typing.Optional[inspect.FullArgSpec] = None

    @root_validator(pre=True)
    def _parse_spec(cls, values: dict):
        values['spec'] = inspect.getfullargspec(values['callback'])
        return values


class Handler(pydantic.BaseModel):
    callback: HandlerCallback
    filters: typing.Tuple[BaseFilter, ...]

    class Config:
        arbitrary_types_allowed = True

    async def check(self, event: types.events.RoomEvent, client: 'AiomatrixClient') -> bool:
        for event_filter in self.filters:
            check_result = await event_filter.check(event, client)
            if not check_result:
                return False
        return True



import datetime

from pydantic import BaseModel, Extra, root_validator

from ...utils.mixins import ContextVarMixin


class MatrixObject(BaseModel, ContextVarMixin):
    raw: dict

    class Config:
        allow_mutation = True
        json_encoders = {datetime.datetime: lambda dt: int(dt.timestamp())}
        validate_assignment = True
        extra = Extra.ignore

    @root_validator(pre=True)
    def _parse_raw(cls, values: dict):
        values['raw'] = values.copy()
        return values



from typing import List

import pydantic

from .. import primitives


class ResolveRoomAliasResponse(pydantic.BaseModel):
    room_id: primitives.RoomID
    servers: List[str]


class GetRoomAliasesResponse(pydantic.BaseModel):
    aliases: List[primitives.RoomAlias]



from typing import List, Optional

import aiomatrix.types.events
from . import models


class BasePresenceStorage:
    async def setup(self):
        raise NotImplementedError

    async def get_user_data(
            self, account_id: str, user_id: aiomatrix.types.primitives.UserID
    ) -> Optional[models.PresenceInDB]:
        raise NotImplementedError

    async def is_new_user(self, account_id: str, user_id: aiomatrix.types.primitives.UserID) -> bool:
        raise NotImplementedError

    async def add_new_presence_update(
            self, account_id: str, event: aiomatrix.types.events.modules.presence.PresenceEvent
    ):
        raise NotImplementedError

    async def update_user_presence(
            self, account_id: str, user_id: aiomatrix.types.primitives.UserID,
            event: aiomatrix.types.events.modules.presence.CurrentPresenceStatus
    ):
        raise NotImplementedError

    async def get_unupdated_users(self, account_id: str, timeout: int) -> List[models.PresenceInDB]:
        raise NotImplementedError



import json
import pathlib
import sqlite3
from typing import List, Optional, Type, TypeVar, Union

import pydantic

T = TypeVar('T')


class SqliteDBConn:
    def __init__(self, db_name):
        self.db_name = db_name

    def __enter__(self):
        self.conn = sqlite3.connect(self.db_name)
        self.conn.row_factory = sqlite3.Row
        return self.conn

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.conn.close()
        if exc_val:
            raise


class SqliteConnection:
    def __init__(self, db_path: Union[pathlib.Path, str]):
        self.db = db_path

    def __make_request(
            self,
            sql: str,
            params: Union[tuple, List[tuple]] = None,
            fetch: bool = False,
            mult: bool = False
    ) -> Optional[Union[List[sqlite3.Row], sqlite3.Row]]:
        with SqliteDBConn(self.db) as conn:
            c = conn.cursor()
            try:
                if params is not None:
                    if isinstance(params, list):
                        c.executemany(sql, params)
                    else:
                        c.execute(sql, params)
                else:
                    c.execute(sql)
            except Exception as e:
                print(e)
            if fetch:
                if mult:
                    r = c.fetchall()
                else:
                    r = c.fetchone()
                return r
            else:
                conn.commit()

    @staticmethod
    def _convert_to_model(data: Optional[sqlite3.Row], model: Type[T]) -> Optional[T]:
        if data is not None:
            dict_data = dict(data)
            for k, v in dict_data.items():
                if isinstance(v, str):
                    try:
                        dict_data[k] = json.loads(v)
                    except ValueError:
                        pass
            try:
                model(**dict_data)
            except pydantic.ValidationError as e:
                print(dict_data, e)
            return model(**dict_data)
        else:
            return None

    def _make_request(
            self,
            sql: str,
            params: Union[tuple, List[tuple]] = None,
            fetch: bool = False,
            mult: bool = False,
            model_type: Type[T] = None
    ) -> Optional[Union[List[T], T]]:
        raw = self.__make_request(sql, params, fetch, mult)
        if raw is None:
            if mult:
                return []
            else:
                return None
        else:
            if mult:
                if model_type is not None:
                    return [self._convert_to_model(i, model_type) for i in raw]
                else:
                    return [i for i in raw]
            else:
                if model_type is not None:
                    return self._convert_to_model(raw, model_type)
                else:
                    return raw



from typing import List

import pydantic

from ..events import RoomEvent, RoomStateEvent


class RoomMessagesResponse(pydantic.BaseModel):
    start: str
    end: str
    chunk: List[RoomEvent]
    state: List[RoomStateEvent]



import datetime
from typing import Optional, Union

from pydantic import BaseModel, Field

from .base import MatrixObject
from .room_state_events import RoomStateContent
from .. import primitives
from ..misc import RoomMessageEventMsgTypesEnum
from ...utils.mixins import ContextClientMixin


class BasicEvent(MatrixObject):
    content: primitives.EventContent
    type: str


class UnsignedData(BaseModel):
    age: Optional[int]
    redacted_because: Optional[dict]  # exists if event was redacted while client was offline
    transcation_id: Optional[str]


class RoomEvent(BasicEvent, ContextClientMixin):
    event_id: primitives.EventID
    sender: primitives.UserID
    timestamp: datetime.datetime = Field(..., alias='origin_server_ts')
    redacts: Optional[primitives.EventID]
    unsigned: Optional[UnsignedData]
    room_id: Optional[primitives.RoomID]  # None only in /sync


class RoomStateEvent(RoomEvent):
    content: RoomStateContent
    state_key: str
    prev_content: Optional[primitives.EventContent]


class RelationshipToEventData(MatrixObject):
    rel_type: str  # add enum for realtionship test_types
    event_id: primitives.EventID


class ReplyToRelationshipData(MatrixObject):
    event_id: primitives.EventID


class ReplyToData(MatrixObject):
    reply_data: ReplyToRelationshipData = Field(..., alias='m.in_reply_to')


class RelationshipMixin(MatrixObject):
    relationship: Optional[Union[RelationshipToEventData, ReplyToData]] = Field(None, alias='m.relates_to')


class BaseMessageEventContent(MatrixObject):
    body: Optional[str]
    msgtype: Union[RoomMessageEventMsgTypesEnum, str]



from typing import List, Optional

import pydantic


class RoomCanonicalAliasContent(pydantic.BaseModel):
    alias: Optional[str]
    alt_aliases: Optional[List[str]]


# class RoomCanonicalAlias(RoomStateEvent):
#     content: RoomCanonicalAliasContent



from typing import Any, List

import pydantic

from aiomatrix.types import primitives


class ToDeviceEvent(pydantic.BaseModel):
    content: Any
    sender: primitives.UserID
    type: str


class ToDevice(pydantic.BaseModel):
    events: List[ToDeviceEvent]



import asyncio
import time
from typing import Callable, List, Union

import pydantic

from .filters import BaseFilter
from .observers import MatrixEventsObserver
from .storage import StorageRepo
from .. import exceptions, loggers, types
from ..client import AiomatrixClient

log = loggers.dispatcher


def _clear_filters(filters: List[BaseFilter]) -> List[BaseFilter]:
    """
    removing repeating filters from list
    """
    result = {}
    for i in filters:
        result[i.filter_id] = i
    return list(result.values())


class AiomatrixDispatcher:
    def __init__(self, clients: List[AiomatrixClient], data_storage: StorageRepo):
        self.storage: StorageRepo = data_storage
        self._clients: List[AiomatrixClient] = clients
        self._state_observer: MatrixEventsObserver = MatrixEventsObserver()
        self._generic_observer: MatrixEventsObserver = MatrixEventsObserver()
        self._redaction_observer: MatrixEventsObserver = MatrixEventsObserver()
        self._lock = asyncio.Lock()

    def register_state_handler(
            self, callback: Callable, *filters: BaseFilter
    ):
        self._state_observer.register(callback, *filters)

    def register_generic_handler(
            self, callback: Callable, *filters: BaseFilter
    ):
        self._generic_observer.register(callback, *filters)

    def register_message_handler(
            self, callback: Callable, *filters: BaseFilter
    ):
        from .filters import builtin
        self._generic_observer.register(
            callback,
            builtin.Incoming(),
            builtin.EventType(event_types=[types.misc.RoomEventTypesEnum.room_message]),
            builtin.MessageType(msg_type=[types.misc.RoomMessageEventMsgTypesEnum.text]),
            *filters
        )

    def register_edited_message_handler(
            self, callback: Callable, *filters: BaseFilter
    ):
        from .filters import builtin
        self._generic_observer.register(
            callback,
            builtin.Incoming(),
            builtin.EventType(event_types=[types.misc.RoomEventTypesEnum.room_message]),
            builtin.MessageType(msg_type=[types.misc.RoomMessageEventMsgTypesEnum.text]),
            builtin.IsEditedMessage(is_edited=True),
            *filters
        )

    def register_redaction_handler(
            self, callback: Callable, *filters: BaseFilter
    ):
        from .filters import builtin
        self._redaction_observer.register(
            callback,
            builtin.EventType([types.misc.RoomEventTypesEnum.redaction]),
            *filters
        )

    async def _save_events_in_db(self, client: AiomatrixClient, events: List[types.events.RoomEvent]):
        batch_size = 500
        parts = [
            events[i:i + batch_size] for i in range(
                0, len(events), batch_size
            )
        ]
        for i, part in enumerate(parts):
            events_in_db = await self.storage.events_repo.are_new_events(
                client.me, list(map(lambda x: x.event_id, part))
            )
            await self.storage.events_repo.insert_new_events_batch(
                client.me, list(filter(lambda x: not events_in_db[x.event_id], part))
            )

    @staticmethod
    def _parse_event(
            event: Union[types.events.RoomEvent, types.events.RoomStateEvent],
            ignore_errors: bool = True
    ) -> Union[types.events.RoomEvent, types.events.RoomMessageEvent, types.events.RoomStateEvent]:
        if isinstance(event, types.events.RoomEvent):
            # if event.unsigned.redacted_because is not None:
            #     # dirty hack to get to original event filed names
            #     event = types.events.RoomRedactionEvent(**json.loads(event.json(by_alias=True)))
            #     event.unsigned.redacted_because = test_types.events.RoomRedactionEvent(**event.unsigned.redacted_because)
            #     return event
            if event.type == types.misc.RoomEventTypesEnum.room_message:
                if event.content is not None:
                    try:
                        event = types.events.RoomMessageEvent.parse_obj(event.dict(by_alias=True))
                    except pydantic.ValidationError as e:
                        if ignore_errors:
                            return event
                        raise e
                    else:
                        return event
            elif event.type == types.misc.RoomEventTypesEnum.reaction:
                if event.content:
                    event.content = types.events.relationships.ReactionRelationshipContent.parse_obj(event.content)
        elif isinstance(event, types.events.RoomStateEvent):
            log.debug(f'received state event: {event}')
        return event

    async def process_sync(
            self, client: AiomatrixClient, state: types.responses.SyncResponse, process_joined_rooms: bool = True,
            process_invited_rooms: bool = True, process_left_rooms: bool = True, process_presence: bool = False
    ):
        """
        main entry point for events to be processed
        :param client:
        :param state:
        :param process_joined_rooms:
        :param process_invited_rooms:
        :param process_left_rooms:
        :param process_presence:
        :return:
        """
        AiomatrixClient.set(client)
        if state.rooms:
            if process_joined_rooms and state.rooms.join:
                for room_id in state.rooms.join:
                    room_data = state.rooms.join[room_id]
                    if room_data.state.events:
                        for event in room_data.state.events:
                            event.room_id = room_id
                        await self._save_events_in_db(client, room_data.state.events)
                for room_id in state.rooms.join:
                    room_data = state.rooms.join[room_id]
                    parsed_events = []
                    if room_data.state.events:
                        for event in room_data.state.events:
                            event.room_id = room_id
                            parsed_events.append(self._parse_event(event))
                    for event in room_data.timeline.events:
                        event.room_id = room_id
                        parsed_events.append(self._parse_event(event))
                    for event in parsed_events:
                        if isinstance(event, types.events.RoomStateEvent):
                            await self._state_observer.trigger(event, client)
                        elif any(isinstance(event, i) for i in [types.events.RoomEvent, types.events.RoomMessageEvent]):
                            if event.type != types.misc.RoomEventTypesEnum.redaction:
                                await self._generic_observer.trigger(event, client)
                            else:
                                await self._redaction_observer.trigger(event, client)
                    await self._save_events_in_db(client, room_data.timeline.events)
            if process_invited_rooms:
                if state.rooms.invite:
                    log.debug(f'{state.rooms.invite=}')
            if process_left_rooms:
                if state.rooms.leave:
                    log.debug(f'{state.rooms.leave=}')
        if process_presence:
            if state.presence:
                for event in state.presence.events:
                    await self.storage.presence_storage.add_new_presence_update(client.me, event)
            unupdated_users = await self.storage.presence_storage.get_outdated_users_data(client.me, 600)
            for user in unupdated_users:
                try:
                    user_presence = await client.presence_api.get_user_presence(user.user_id)
                except exceptions.Forbidden:
                    log.error(f'couldn\'t get presence status for {user.user_id}: forbidden')
                except Exception as e:
                    log.exception(f'error getting presence status: {e=}')
                else:
                    await self.storage.presence_storage.update_user_presence(client.me, user.user_id, user_presence)

    async def run_polling(
            self, ignore_errors: bool = True, timeout: int = 10, sleep: float = 0.1, track_presence: bool = False
    ):
        async with self._lock:  # to prevent multiple pollings
            await self.storage.setup()
            coro_list = []
            for i in self._clients:
                coro_list.append(
                    self._run_bot_polling(i, ignore_errors, timeout, sleep, track_presence)
                )
            await asyncio.gather(*coro_list)

    async def _run_bot_polling(
            self, client: AiomatrixClient, ignore_errors: bool = True, timeout: int = 10,
            sleep: float = 0.1, track_presence: bool = False
    ):
        while True:
            await client.login()
            next_batch = await self.storage.internal_repo.get_last_sync_token(client.me)
            try:
                state = await client.sync_api.sync(
                    full_state=False,
                    since=next_batch,
                    timeout=timeout * 1000
                )
            except (exceptions.MissingToken, exceptions.UnknownToken) as e:
                log.exception(f'login error {e=} for client {client.me}')
                await client.auth_api.logout()
                await client.login()
                if not client.auth_api.is_logged_in:
                    log.exception(f'couldn\'t recover from login error for client {client.me}')
                    break
            except Exception as e:
                log.exception(f'sync error {e=} for client {client.me}')
                if not ignore_errors:
                    raise e
            else:
                t = time.time()
                await self.process_sync(client, state, process_joined_rooms=True, process_presence=track_presence)
                log.debug(f'sync for client {client.me} proceeded in {time.time() - t}')
                await self.storage.internal_repo.set_last_sync_token(client.me, state.next_batch)
            finally:
                await asyncio.sleep(sleep)



import pathlib
from typing import Optional, Union

from . import BaseInternalDataStorage, engines, models


class SqliteInternalDataStorage(BaseInternalDataStorage):
    def __init__(self, db_path: Union[pathlib.Path, str]):
        self.db_mngr = engines.SqliteInternalDataStorageEngine(db_path)

    async def setup(self):
        await self.db_mngr.create_internal_data_table()

    async def get_last_sync_token(self, account_id: str) -> Optional[models.InternalDataPair]:
        db_res = await self.db_mngr.get_data_by_key(account_id, 'sync_token')
        return db_res

    async def set_last_sync_token(self, account_id: str, token: str):
        if await self.get_last_sync_token(account_id) is not None:
            await self.db_mngr.set_data_by_key(account_id, 'sync_token', token)
        else:
            await self.db_mngr.insert_data(account_id, 'sync_token', token)



from typing import Dict, List, Optional

from aiomatrix import types
from . import models


class BaseEventStorage:
    async def setup(self):
        raise NotImplementedError
    
    async def get_event_data(
            self, event_id: types.primitives.EventID
    ) -> Optional[models.EventInDB]:
        raise NotImplementedError

    async def is_new_event(self, event_id: types.primitives.EventID) -> bool:
        raise NotImplementedError

    async def get_events_by_room(
            self, room_id: types.primitives.RoomID
    ) -> List[models.EventInDB]:
        raise NotImplementedError

    async def get_multiple_events_data(
            self, event_ids: List[types.primitives.EventID]
    ) -> Dict[str, models.EventInDB]:
        raise NotImplementedError

    async def insert_new_event(self, event: types.events.RoomEvent):
        raise NotImplementedError

    async def insert_new_events_batch(self, events: List[types.events.RoomEvent]):
        raise NotImplementedError

    async def get_account_seen_events(self, account_id: str) -> List[models.SeenEvent]:
        raise NotImplementedError

    async def check_seen_events(
            self, account_id: str, event_ids: List[types.primitives.EventID]
    ) -> List[models.SeenEvent]:
        raise NotImplementedError

    async def add_new_seen_event(self, account_id: str, event_id: types.primitives.EventID):
        raise NotImplementedError

    async def add_new_seen_events_batch(self, account_id: str, event_ids: List[types.primitives.EventID]):
        raise NotImplementedError



from . import modules
from .auth import AuthAPI
from .capabilities_negotiation import CapabilitiesAPI
from .listing_rooms import ListingRoomsAPI
from .room_membership import RoomMembershipAPI
from .rooms import RoomsAPI
from .syncing import SyncingAPI



from typing import Optional

import pydantic
from pydantic import Field

from ... import primitives


class PreviousRoom(pydantic.BaseModel):
    room_id: primitives.RoomID
    event_id: primitives.EventID


class RoomCreateContent(pydantic.BaseModel):
    creator: primitives.UserID
    federate: bool = Field(True, alias='m.federate')
    room_version: str = '1'
    predecessor: Optional[PreviousRoom]


# class RoomCreate(RoomStateEvent):
#     content: RoomCreateContent



from typing import List, Optional

import pydantic

from .. import primitives
from ..misc import RoomVisiblityEnum


class RoomVisibilityResponse(pydantic.BaseModel):
    visibility: RoomVisiblityEnum


class ServerPublicRoomsChunk(pydantic.BaseModel):
    aliases: List[primitives.RoomAlias] = []
    canonical_alias: Optional[str]
    name: Optional[str]
    num_joined_members: int
    room_id: primitives.RoomID
    topic: Optional[str]
    world_readable: bool
    guest_can_join: bool
    avatar_url: Optional[str]


class ServerPublicRoomsResponse(pydantic.BaseModel):
    chunk: List[ServerPublicRoomsChunk]
    next_batch: Optional[str]
    prev_batch: Optional[str]
    total_room_count_estimate: Optional[int]



from typing import Dict

import pydantic
from pydantic import Field

from ..misc import RoomStabilityEnum


class ChangePasswordCapability(pydantic.BaseModel):
    enabled: bool


class RoomVersionsCapability(pydantic.BaseModel):
    default: str
    available: Dict[str, RoomStabilityEnum]


class ServerCapabilities(pydantic.BaseModel):
    change_password: ChangePasswordCapability = Field(None, alias='m.change_password')
    room_versions: RoomVersionsCapability = Field(None, alias='m.room_versions')


class ServerCapabilitiesResponse(pydantic.BaseModel):
    capabilities: ServerCapabilities



import io
import os.path
import pathlib
from abc import ABC, abstractmethod
from typing import AsyncGenerator, AsyncIterator, Optional, Union

import aiofiles

DEFAULT_CHUNK_SIZE = 64 * 1024  # 64 KB


class InputFile(ABC):
    def __init__(self, filename: Optional[str] = None, chunk_size: int = DEFAULT_CHUNK_SIZE):
        self.filename = filename
        self.chunk_size = chunk_size

    @abstractmethod
    async def read(self, chunk_size: int) -> AsyncGenerator[bytes, None]:
        yield b""

    async def __aiter__(self) -> AsyncIterator[bytes]:
        async for chunk in self.read(self.chunk_size):
            yield chunk


class FSInputFile(InputFile):
    def __init__(
            self, path: Union[pathlib.Path, str], filename: Optional[str] = None, chunk_size: int = DEFAULT_CHUNK_SIZE
    ):
        if filename is None:
            filename = os.path.basename(path)
        super().__init__(filename, chunk_size)
        self.path = path

    async def read(self, chunk_size: int) -> AsyncGenerator[bytes, None]:
        async with aiofiles.open(self.path, 'rb') as f:
            chunk = await f.read(chunk_size)
            while chunk:
                yield chunk
                chunk = await f.read(chunk_size)


class BufferedInputFile(InputFile):
    def __init__(self, data: Union[io.BytesIO, bytes], filename: Optional[str] = None, chunk_size: int = DEFAULT_CHUNK_SIZE):
        super().__init__(filename, chunk_size)
        self.data = data if isinstance(data, bytes) else data.getvalue()

    async def read(self, chunk_size: int) -> AsyncGenerator[bytes, None]:
        buffer = io.BytesIO(self.data)
        chunk = buffer.read(chunk_size)
        while chunk:
            yield chunk
            chunk = buffer.read(chunk_size)



from typing import Optional

from aiomatrix import types
from . import raw_api


class AuthAPI:
    def __init__(self, raw_api_client: raw_api.RawAPI):
        self._raw_api = raw_api_client

    async def get_login_types(self) -> types.responses.SupportedLoginTypes:
        r = await self._raw_api.make_request(
            'GET', '_matrix/client/r0/login', model_type=types.responses.SupportedLoginTypes
        )
        return r

    async def password_login(
            self, login: str, password: str, device_id: Optional[str] = None
    ) -> types.responses.LoginResponse:
        payload = {
            'data': {
                'type':                        'm.login.password',
                'identifier':                  {
                    'type': 'm.id.user',
                    'user': login
                },
                'password':                    password,
                'initial_device_display_name': 'aiomatrix-bot'
            }
        }
        if device_id is not None:
            payload['data']['device_id'] = device_id
        r = await self._raw_api.make_request(
            'POST', '_matrix/client/r0/login', model_type=types.responses.LoginResponse, **payload
        )
        return r

    def set_access_token(self, access_token: str):
        self._raw_api.access_token = access_token

    @property
    def is_logged_in(self):
        return self._raw_api.is_authorized

    async def logout(self):
        await self._raw_api.make_request('POST', '_matrix/client/r0/logout')
        self._raw_api.access_token = ''

    async def logout_all(self):
        await self._raw_api.make_request('POST', '_matrix/client/r0/logout/all')

    async def whoami(self) -> types.responses.WhoAmIResponse:
        r = await self._raw_api.make_request(
            'GET', '_matrix/client/r0/account/whoami', model_type=types.responses.WhoAmIResponse
        )
        return r



import typing
from typing import List, Union

from aiomatrix import types
from .base import BaseFilter

if typing.TYPE_CHECKING:
    from aiomatrix import AiomatrixClient


class RoomID(BaseFilter):
    def __init__(self, room_ids: List[str]):
        super().__init__()
        self.filter_id: str = 'room_id'
        self.room_ids = set(room_ids)

    async def check(self, event: types.events.RoomEvent, client: 'AiomatrixClient') -> bool:
        return event.room_id in self.room_ids


class EventType(BaseFilter):
    def __init__(self, event_types: List[Union[str, types.misc.RoomEventTypesEnum]]):
        super().__init__()
        self.filter_id: str = 'event_type'
        self.event_types = set(map(lambda x: f'{x}', event_types))

    async def check(self, event: types.events.RoomEvent, client: 'AiomatrixClient') -> bool:
        return event.type in self.event_types


class MessageType(BaseFilter):
    def __init__(self, msg_type: List[Union[str, types.misc.RoomMessageEventMsgTypesEnum]]):
        super().__init__()
        self.filter_id: str = 'message_type'
        self.message_types = set(map(lambda x: f'{x}', msg_type))

    async def check(self, event: types.events.RoomMessageEvent, client: 'AiomatrixClient') -> bool:
        if event.content is not None:
            if isinstance(event.content, types.events.modules.instant_messaging.NewContent):
                return event.content.new_content.msgtype in self.message_types
            return event.content.msgtype in self.message_types
        return False


class SenderID(BaseFilter):
    def __init__(self, sender_ids: List[str]):
        super().__init__()
        self.filter_id: str = 'sender_id'
        self.sender_ids = set(sender_ids)

    async def check(self, event: types.events.RoomEvent, client: 'AiomatrixClient') -> bool:
        return event.sender in self.sender_ids


class Incoming(BaseFilter):
    """
    only incoming events
    """

    def __init__(self):
        super().__init__()
        self.filter_id: str = 'message_direction'

    async def check(self, event: types.events.RoomEvent, client: 'AiomatrixClient') -> bool:
        return event.sender != client.me.user_id


class Outgoing(BaseFilter):
    """
    only outgoing events
    """

    def __init__(self):
        super().__init__()
        self.filter_id: str = 'message_direction'

    async def check(self, event: types.events.RoomEvent, client: 'AiomatrixClient') -> bool:
        return event.sender == client.me.user_id


class Text(BaseFilter):
    def __init__(self, texts: List[str], case_insensitive: bool = True):
        super().__init__()
        self.filter_id: str = 'text'
        self.case_insensitive = case_insensitive
        if self.case_insensitive:
            self.allowed_texts = map(lambda x: x.lower(), texts)
        else:
            self.allowed_texts = texts
        self.allowed_texts = set(self.allowed_texts)

    async def check(self, event: types.events.RoomMessageEvent, client: 'AiomatrixClient') -> bool:
        event_txt = event.content.body
        if self.case_insensitive:
            event_txt = event_txt.lower()
        return any((i in event_txt for i in self.allowed_texts))


class Command(Text):
    def __init__(self, commands: List[str], prefix: str = '!'):
        super().__init__(list(map(lambda x: f'{prefix}{x}', commands)), True)


class IsEditedMessage(BaseFilter):
    def __init__(self, is_edited: bool = True):
        super().__init__()
        self.filter_id: str = 'is_edited_message'
        self.is_edited = is_edited

    async def check(self, event: types.events.RoomMessageEvent, client: 'AiomatrixClient') -> bool:
        if event.content is not None:
            return isinstance(event.content, types.events.modules.instant_messaging.NewContent) == self.is_edited



from typing import List, Optional, Union

import aiomatrix.types.events.modules.presence.presence
from . import BasePresenceStorage, models


# TODO: remove account_id from storage
class PresenceRepo:
    def __init__(self, storage: BasePresenceStorage):
        self.storage = storage

    async def setup(self):
        await self.storage.setup()

    async def get_user_data(
            self, account_id: Union[str, aiomatrix.types.responses.WhoAmIResponse],
            user_id: aiomatrix.types.primitives.UserID
    ) -> Optional[models.PresenceInDB]:
        if isinstance(account_id, aiomatrix.types.responses.WhoAmIResponse):
            account_id = account_id.user_id
        return await self.storage.get_user_data(account_id, user_id)

    async def is_new_user(
            self, account_id: Union[str, aiomatrix.types.responses.WhoAmIResponse],
            user_id: aiomatrix.types.primitives.UserID
    ) -> bool:
        if isinstance(account_id, aiomatrix.types.responses.WhoAmIResponse):
            account_id = account_id.user_id
        return await self.storage.is_new_user(account_id, user_id)

    async def add_new_presence_update(
            self, account_id: Union[str, aiomatrix.types.responses.WhoAmIResponse],
            event: aiomatrix.types.events.modules.presence.PresenceEvent
    ):
        if isinstance(account_id, aiomatrix.types.responses.WhoAmIResponse):
            account_id = account_id.user_id
        return await self.storage.add_new_presence_update(account_id, event)

    async def update_user_presence(
            self, account_id: Union[str, aiomatrix.types.responses.WhoAmIResponse],
            user_id: aiomatrix.types.primitives.UserID,
            event: aiomatrix.types.events.modules.presence.CurrentPresenceStatus
    ):
        if isinstance(account_id, aiomatrix.types.responses.WhoAmIResponse):
            account_id = account_id.user_id
        return await self.storage.update_user_presence(account_id, user_id, event)

    async def get_outdated_users_data(
            self, account_id: Union[str, aiomatrix.types.responses.WhoAmIResponse], timeout: int = 60
    ) -> List[models.PresenceInDB]:
        if isinstance(account_id, aiomatrix.types.responses.WhoAmIResponse):
            account_id = account_id.user_id
        return await self.storage.get_unupdated_users(account_id, timeout)



from __future__ import annotations

import typing

from pydantic import validator

from .base_room_events import RoomEvent
from .modules import instant_messaging
from ..misc import RoomMessageEventMsgTypesEnum

if typing.TYPE_CHECKING:
    from ..responses import SentEventResponse


class RoomMessageEvent(RoomEvent):
    content: typing.Optional[
        typing.Union[
            instant_messaging.NewContent, instant_messaging.LocationContent, instant_messaging.TextContent,
            instant_messaging.NoticeContent, instant_messaging.AudioContent, instant_messaging.EmoteContent,
            instant_messaging.FileContent, instant_messaging.ImageContent, instant_messaging.VideoContent,
        ]
    ]

    @validator('content')
    def _fix_typing(
            cls,
            v: typing.Optional[
                typing.Union[
                    instant_messaging.NewContent, instant_messaging.LocationContent, instant_messaging.TextContent,
                    instant_messaging.NoticeContent, instant_messaging.AudioContent, instant_messaging.EmoteContent,
                    instant_messaging.FileContent, instant_messaging.ImageContent, instant_messaging.VideoContent
                ]
            ]
    ):
        if v is not None:
            supported_msgtypes = {
                RoomMessageEventMsgTypesEnum.audio:    instant_messaging.AudioContent,
                RoomMessageEventMsgTypesEnum.emote:    instant_messaging.EmoteContent,
                RoomMessageEventMsgTypesEnum.file:     instant_messaging.FileContent,
                RoomMessageEventMsgTypesEnum.image:    instant_messaging.ImageContent,
                RoomMessageEventMsgTypesEnum.location: instant_messaging.LocationContent,
                RoomMessageEventMsgTypesEnum.notice:   instant_messaging.NoticeContent,
                RoomMessageEventMsgTypesEnum.text:     instant_messaging.TextContent,
                RoomMessageEventMsgTypesEnum.video:    instant_messaging.VideoContent,
            }
            raw_content = v.raw
            if 'm.new_content' in raw_content and isinstance(v.raw['m.new_content'], dict):
                if 'msgtype' not in raw_content['m.new_content']:
                    raise ValueError('not found message type in new content')
                if raw_content['m.new_content']['msgtype'] not in supported_msgtypes:
                    raise ValueError(f'unknown new content type: {raw_content["msgtype"]["m.new_content"]}')
                content = instant_messaging.NewContent.parse_obj(raw_content)
            else:
                if 'msgtype' not in raw_content:
                    return None
                if raw_content['msgtype'] not in supported_msgtypes:
                    raise ValueError(f'unknown message type: {raw_content["msgtype"]}')
                content = supported_msgtypes[raw_content['msgtype']](**raw_content)
            return content

    async def send_reaction(self, key: str) -> 'SentEventResponse':
        return await self.client.instant_messaging_api.send_reaction(self.room_id, self.event_id, key)

    async def send_typing(self, timeout: int = 5000):
        await self.client.typing_notifications_api.send_typing(self.client.me.user_id, self.room_id, timeout)

    async def send_notice(self, text: str) -> 'SentEventResponse':
        return await self.client.instant_messaging_api.send_notice(self.room_id, text)

    async def answer_notice(self, text: str) -> 'SentEventResponse':
        return await self.client.instant_messaging_api.send_notice(self.room_id, text, reply_to=self.event_id)



from typing import Optional

import pydantic


class RoomRedactionContent(pydantic.BaseModel):
    reason: Optional[str]


# class RoomRedactionEvent(RoomEvent):  # redaction of already sent event
#     content: RoomRedactionContent



from __future__ import annotations

from typing import Optional

from .image import BasicImageInfo
from ..e2ee import EncryptedFile
from ...base_room_events import BaseMessageEventContent


class LocationInfo(BasicImageInfo):
    thumbnail_url: Optional[str]
    thumnail_file: Optional[EncryptedFile]
    thumbnail_info: Optional[BasicImageInfo]


class LocationContent(BaseMessageEventContent):
    info: Optional[LocationInfo]
    geo_uri: str



from . import BaseEventStorage, BaseInternalDataStorage, BasePresenceStorage
from .internal_data import InternalDataRepo
from .presence import PresenceRepo
from .room_events import EventsRepo


class StorageRepo:
    def __init__(
            self, internal_storage: BaseInternalDataStorage, events_storage: BaseEventStorage,
            presence_storage: BasePresenceStorage
    ):
        self.internal_repo = InternalDataRepo(internal_storage)
        self.events_repo = EventsRepo(events_storage)
        self.presence_storage = PresenceRepo(presence_storage)

    async def setup(self):
        await self.internal_repo.setup()
        await self.events_repo.setup()
        await self.presence_storage.setup()



from __future__ import annotations

from typing import Optional

from .basic_file_info import BasicFileInfo
from ..e2ee import EncryptedFile
from ...base_room_events import BaseMessageEventContent


class BasicImageInfo(BasicFileInfo):
    h: int
    w: int


class FileInfo(BasicImageInfo):
    thumbnail_url: Optional[str]
    thumnail_file: Optional[EncryptedFile]
    thumbnail_info: Optional[BasicImageInfo]


class ImageContent(BaseMessageEventContent):
    info: Optional[FileInfo]
    url: Optional[str]
    file: Optional[EncryptedFile]



from typing import Optional

from aiomatrix import types
from . import raw_api


class ListingRoomsAPI:
    def __init__(self, raw_api_client: raw_api.RawAPI):
        self._raw_api = raw_api_client

    async def get_room_visibility(self, room_id: types.primitives.RoomID) -> types.responses.RoomVisibilityResponse:
        r = await self._raw_api.make_request(
            'GET', f'_matrix/client/r0/directory/list/room/{room_id}', model_type=types.responses.RoomVisibilityResponse
        )
        return r

    async def set_room_visibility(self, room_id: types.primitives.RoomID, visibility: types.misc.RoomVisiblityEnum):
        payload = {
            'data': {
                'visibility': f'{visibility}'
            }
        }
        await self._raw_api.make_request(
            'PUT', f'_matrix/client/r0/directory/list/room/{room_id}', model_type=None, **payload
        )

    async def get_server_public_rooms(
            self, limit: Optional[int] = None, since: Optional[str] = None, server: Optional[str] = None
    ) -> types.responses.ServerPublicRoomsResponse:
        payload = {'params': {}}
        if limit is not None:
            payload['params']['limit'] = limit
        if since is not None:
            payload['params']['since'] = since
        if server is not None:
            payload['params']['server'] = server
        r = await self._raw_api.make_request(
            'GET', f'_matrix/client/r0/publicRooms', model_type=types.responses.ServerPublicRoomsResponse, **payload
        )
        return r

    async def get_server_all_public_rooms(
            self, server: Optional[str] = None
    ) -> types.responses.ServerPublicRoomsResponse:
        result = types.responses.ServerPublicRoomsResponse(chunk=[])
        step = 250
        since = None
        while True:
            server_rooms_chunk = await self.get_server_public_rooms(step, since, server)
            result.total_room_count_estimate = server_rooms_chunk.total_room_count_estimate
            result.chunk.extend(server_rooms_chunk.chunk)
            if server_rooms_chunk.next_batch is None:
                break
            else:
                since = server_rooms_chunk.next_batch
        return result



import logging

logging.basicConfig(
    format='%(asctime)s - %(levelname)s - %(name)s - %(message)s'
)

dispatcher = logging.getLogger('aiomatrix.dispatcher')
client = logging.getLogger('aiomatrix.client')



import abc
import typing
import uuid

from aiomatrix import types

if typing.TYPE_CHECKING:
    from aiomatrix import AiomatrixClient


class BaseFilter(metaclass=abc.ABCMeta):
    def __init__(self):
        self.filter_id: str = f'{uuid.uuid4()}'

    @abc.abstractmethod
    async def check(self, event: types.events.BasicEvent, client: 'AiomatrixClient') -> bool:
        raise NotImplementedError



import uuid
from typing import List, Optional

from aiomatrix import types
from . import raw_api


class SyncingAPI:
    def __init__(self, raw_api_client: raw_api.RawAPI):
        self._raw_api = raw_api_client
        self.state = dict

    async def sync(
            self, filter_id: Optional[str] = None, since: Optional[str] = None, full_state: bool = False,
            set_presence: types.misc.PresenceEnum = types.misc.PresenceEnum.online,
            timeout: int = 0
    ) -> types.responses.SyncResponse:
        payload = {'params': {}}
        if filter_id is not None:
            payload['params']['filter'] = filter_id
        if since is not None:
            payload['params']['since'] = since
        if full_state is not None:
            payload['params']['full_state'] = full_state.__str__().lower()
        if set_presence is not None:
            payload['params']['set_presence'] = f'{set_presence}'
        if timeout is not None:
            payload['params']['timeout'] = timeout
        r = await self._raw_api.make_request(
            'GET', '_matrix/client/r0/sync', model_type=types.responses.SyncResponse, **payload
        )
        return r

    async def get_room_event(
            self, room_id: types.primitives.RoomID, event_id: types.primitives.EventID
    ) -> types.events.RoomEvent:
        r = await self._raw_api.make_request(
            'GET', f'_matrix/client/r0/rooms/{room_id}/event/{event_id}', model_type=types.events.RoomEvent
        )
        return r

    async def get_room_state(self, room_id: types.primitives.RoomID) -> List[types.events.RoomEvent]:
        r = await self._raw_api.make_request(
            'GET', f'_matrix/client/r0/rooms/{room_id}/room_events', model_type=types.events.RoomEvent
        )
        return r

    async def get_room_members(
            self, room_id: types.primitives.RoomID, at: Optional[str] = None,
            membership: Optional[types.events.RoomMemberMembershipEnum] = None,
            not_membership: Optional[types.events.RoomMemberMembershipEnum] = None
    ) -> types.responses.RoomMembersResponse:
        payload = {'params': {}}
        if at is not None:
            payload['params']['at'] = at
        if membership is not None:
            payload['params']['membership'] = f'{membership}'
        if not_membership is not None:
            payload['params']['not_membership'] = f'{not_membership}'
        r = await self._raw_api.make_request(
            'GET', f'_matrix/client/r0/rooms/{room_id}/members', model_type=types.responses.RoomMembersResponse,
            **payload
        )
        return r

    async def get_room_messages(
            self, room_id: types.primitives.RoomID, from_token: str, to: Optional[str] = None, direction: str = 'f',
            limit: int = 10, filter_id: Optional[str] = None
    ) -> types.responses.RoomMessagesResponse:
        payload = {'params': {'from': from_token}}
        if to is not None:
            payload['params']['to'] = to
        if direction is not None:
            payload['params']['dir'] = direction
        if limit is not None:
            payload['params']['limit'] = limit
        if filter_id is not None:
            payload['params']['filter'] = filter_id
        r = await self._raw_api.make_request(
            'GET', f'_matrix/client/r0/rooms/{room_id}/messages', model_type=types.responses.RoomMessagesResponse,
            **payload
        )
        return r

    async def send_message_event(
            self, room_id: types.primitives.RoomID, event_type: types.misc.RoomEventTypesEnum, content: dict
    ) -> types.responses.SentEventResponse:
        payload = {
            'data': content
        }
        txn_id = f'{uuid.uuid4()}'
        r = await self._raw_api.make_request(
            'PUT', f'_matrix/client/r0/rooms/{room_id}/send/{event_type}/{txn_id}',
            model_type=types.responses.SentEventResponse, **payload
        )
        return r



from typing import Dict

import pydantic


class RoomPowerLevelNotifications(pydantic.BaseModel):
    room: int = 50


class RoomPowerLevelContent(pydantic.BaseModel):
    ban: int = 50
    events: Dict[str, int]
    events_default: int = 0
    invite: int = 50
    kick: int = 50
    redact: int = 50
    state_default: int = 50
    users: Dict[str, int]
    users_default: int = 0
    notifications: RoomPowerLevelNotifications = RoomPowerLevelNotifications()


# class RoomPowerLevelEvent(RoomStateEvent):
#     content: RoomPowerLevelContent



from typing import List, Optional

import aiomatrix
from .. import models
from ...base_engines import SqliteConnection


class SqliteEventStorageEngine(SqliteConnection):
    async def create_events_table(self):
        sql = '''
            create table if not exists events
            (
                event_id   TEXT,
                ts         INTEGER,
                room_id    TEXT,
                sender     TEXT,
                event_type TEXT,
                data       TEXT,
                UNIQUE (event_id) ON CONFLICT IGNORE,
                PRIMARY KEY (event_id)
            );
        '''
        self._make_request(sql)

    async def create_seen_events_table(self):
        sql = '''
            create table if not exists seen_events
            (
                account_id TEXT,
                event_id   TEXT,
                UNIQUE (account_id, event_id) ON CONFLICT IGNORE,
                PRIMARY KEY (account_id, event_id)
            );
        '''
        self._make_request(sql)

    async def get_event(self, event_id: aiomatrix.types.primitives.EventID) -> Optional[models.EventInDB]:
        sql = 'SELECT * FROM events WHERE event_id = ?'
        params = (event_id,)
        r = self._make_request(sql, params, fetch=True, model_type=models.EventInDB)
        return r

    async def get_events_batch(
            self, events_ids: List[aiomatrix.types.primitives.EventID]
    ) -> List[models.EventInDB]:
        # noinspection SqlResolve
        sql = f'SELECT * FROM events WHERE event_id IN ({", ".join(["?" for _ in range(len(events_ids))])})'
        params = tuple([*events_ids])
        r = self._make_request(sql, params, fetch=True, mult=True, model_type=models.EventInDB)
        return r

    async def add_new_state_event(self, event: aiomatrix.types.events.RoomEvent):
        sql = 'INSERT INTO events (event_id, room_id, sender, ts, event_type, data) VALUES (?, ?, ?, ?, ?, ?)'
        params = (
            event.event_id, event.room_id, event.sender, int(event.timestamp.timestamp()), event.type,
            event.json(by_alias=True)
        )
        self._make_request(sql, params)

    async def add_new_state_events_batch(self, events: List[aiomatrix.types.events.RoomEvent]):
        sql = 'INSERT INTO events (event_id, room_id, sender, ts, event_type, data) VALUES (?, ?, ?, ?, ?, ?)'
        params = [
            (
                event.event_id, event.room_id, event.sender, int(event.timestamp.timestamp()), event.type,
                event.json(by_alias=True)
            )
            for event in events
        ]
        self._make_request(sql, params)

    async def get_room_events(
            self, room_id: aiomatrix.types.primitives.RoomID
    ) -> List[models.EventInDB]:
        sql = 'SELECT * FROM events WHERE room_id = ? ORDER BY ts'
        params = (room_id,)
        r = self._make_request(sql, params, fetch=True, mult=True, model_type=models.EventInDB)
        return r

    async def get_account_seen_events(self, account_id: str) -> List[models.SeenEvent]:
        sql = 'SELECT * FROM seen_events WHERE account_id = ?'
        params = (account_id,)
        r = self._make_request(sql, params, fetch=True, mult=True, model_type=models.SeenEvent)
        return r

    async def check_seen_events(
            self, account_id: str, events_ids: List[aiomatrix.types.primitives.EventID]
    ) -> List[models.SeenEvent]:
        # noinspection SqlResolve
        sql = f'SELECT * FROM seen_events WHERE account_id = ? AND event_id IN ({", ".join(["?" for _ in range(len(events_ids))])})'
        params = tuple([account_id, *events_ids])
        r = self._make_request(sql, params, fetch=True, mult=True, model_type=models.SeenEvent)
        return r

    async def add_new_seen_event(self, account_id: str, event_id: aiomatrix.types.primitives.EventID):
        sql = 'INSERT INTO seen_events (account_id, event_id) VALUES (?, ?)'
        params = (account_id, event_id)
        self._make_request(sql, params)

    async def add_new_seen_events_batch(self, account_id: str, event_ids: List[aiomatrix.types.primitives.EventID]):
        sql = 'INSERT INTO seen_events (account_id, event_id) VALUES (?, ?)'
        params = [(account_id, event_id) for event_id in event_ids]
        self._make_request(sql, params)



import io
import pathlib
from builtins import int
from typing import AsyncGenerator, BinaryIO, Optional, Union

import aiofiles as aiofiles

from aiomatrix import types, utils
from .. import raw_api


class ContentRepositoryAPI:
    def __init__(self, raw_api_client: raw_api.RawAPI):
        self._raw_api = raw_api_client

    async def get_config(self):
        await self._raw_api.make_request(
            'GET', f'_matrix/media/v3/config',
            model_type=types.responses.ContentRepositoryConfig
        )

    @classmethod
    async def __download_to_binary_io(
            cls, destination: BinaryIO, seek: bool, stream: AsyncGenerator[bytes, None]
    ) -> BinaryIO:
        async for chunk in stream:
            destination.write(chunk)
            destination.flush()
        if seek is True:
            destination.seek(0)
        return destination

    @classmethod
    async def __download_to_file(
            cls, destination: Union[str, pathlib.Path], stream: AsyncGenerator[bytes, None]
    ) -> None:
        async with aiofiles.open(destination, 'wb') as f:
            async for chunk in stream:
                await f.write(chunk)

    async def download_file(
            self, mxc_url: str, destination: Optional[Union[pathlib.Path, str]] = None, chunk_size: int = 65536
    ) -> Optional[BinaryIO]:
        if destination is None:
            destination = io.BytesIO()
        if mxc_url.startswith('mxc://'):
            mxc_url = mxc_url[6:]
        server_name, media_id = mxc_url.split('/')
        stream = self._raw_api.stream_response(f'_matrix/media/r0/download/{server_name}/{media_id}', chunk_size)
        try:
            if isinstance(destination, (str, pathlib.Path)):
                return await self.__download_to_file(destination, stream)
            else:
                return await self.__download_to_binary_io(destination, True, stream)
        finally:
            pass

    async def download_thumbnail(
            self, mxc_url: str, height: int, width: int, method: Union[str, types.misc.ResizingMethodsEnum],
            destination: Optional[Union[pathlib.Path, str]] = None, chunk_size: int = 65536,
    ) -> Optional[BinaryIO]:
        if destination is None:
            destination = io.BytesIO()
        if mxc_url.startswith('mxc://'):
            mxc_url = mxc_url[6:]
        server_name, media_id = mxc_url.split('/')
        payload = {
            'params': {
                'height': height,
                'width':  width,
                'method': method
            }
        }
        stream = self._raw_api.stream_response(
            f'_matrix/media/r0/thumbnail/{server_name}/{media_id}', chunk_size, **payload
        )
        try:
            if isinstance(destination, (str, pathlib.Path)):
                return await self.__download_to_file(destination, stream)
            else:
                return await self.__download_to_binary_io(destination, True, stream)
        finally:
            pass

    async def upload_file(self, file: Union[utils.InputFile, str]) -> types.responses.UploadedFileResponse:
        payload = {
            'data':    file.read(file.chunk_size),
            'params':  {
                           'filename': file.filename
                       } if file.filename is not None else {},
        }
        r = await self._raw_api.make_request(
            'POST', f'_matrix/media/r0/upload',
            model_type=types.responses.UploadedFileResponse, **payload
        )
        return r



from __future__ import annotations

from typing import Optional

from pydantic import BaseModel

from .basic_file_info import BasicFileInfo
from .image import BasicImageInfo
from ..e2ee import EncryptedFile
from ...base_room_events import BaseMessageEventContent


class FileInfo(BasicFileInfo, BaseModel):
    thumbnail_url: Optional[str]
    thumnail_file: Optional[EncryptedFile]
    thumbnail_info: Optional[BasicImageInfo]


class FileContent(BaseMessageEventContent):
    info: Optional[FileInfo]
    url: Optional[str]
    file: Optional[EncryptedFile]



#!/usr/bin/env python3
import pathlib
import re
import sys

from setuptools import find_packages, setup

WORK_DIR = pathlib.Path(__file__).parent

# Check python version
MINIMAL_PY_VERSION = (3, 7)
if sys.version_info < MINIMAL_PY_VERSION:
    raise RuntimeError('aiomatrix works only with Python {}+'.format('.'.join(map(str, MINIMAL_PY_VERSION))))


def get_version():
    """
    Read version
    :return: str
    """
    txt = (WORK_DIR / 'aiomatrix' / '__init__.py').read_text('utf-8')
    try:
        return re.findall(r"^__version__ = '([^']+)'\r?$", txt, re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')


def get_description():
    with open('README.rst', 'r', encoding='utf-8') as f:
        return f.read()


setup(
    name='aiomatrix-py',
    version=get_version(),
    packages=find_packages(exclude=('tests', 'tests.*', 'docs', 'examples.*')),
    url='https://github.com/Forden/aiomatrix',
    license='MIT',
    author='Forden',
    python_requires='>=3.7',
    description='Aiomatrix is a simple library for Matrix Client-Server API',
    long_description=get_description(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Programming Language :: Python',
        'Framework :: AsyncIO',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    install_requires=[
        'aiohttp>=3.7.2,<4.0.0',
        'pydantic==1.8.2',
        'aiofiles==0.7.0'
    ],
    include_package_data=False,
)



from typing import Optional, Union

import aiomatrix.types.responses
from . import BaseInternalDataStorage


class InternalDataRepo:
    def __init__(self, storage: BaseInternalDataStorage):
        self.storage = storage

    async def setup(self):
        await self.storage.setup()

    async def get_last_sync_token(
            self, account_id: Union[str, aiomatrix.types.responses.WhoAmIResponse]
    ) -> Optional[str]:
        if isinstance(account_id, aiomatrix.types.responses.WhoAmIResponse):
            account_id = account_id.user_id
        db_res = await self.storage.get_last_sync_token(account_id)
        if db_res:
            return db_res.data
        return None

    async def set_last_sync_token(self, account_id: Union[str, aiomatrix.types.responses.WhoAmIResponse], token: str):
        if isinstance(account_id, aiomatrix.types.responses.WhoAmIResponse):
            account_id = account_id.user_id
        return await self.storage.set_last_sync_token(account_id, token)



from .content_repository import ContentRepositoryConfig, UploadedFileResponse
from .create_room import CreateRoomResponse
from .listing_rooms import RoomVisibilityResponse, ServerPublicRoomsResponse
from .login_response import LoginResponse
from .login_types import SupportedLoginTypes
from .room_aliases import GetRoomAliasesResponse, ResolveRoomAliasResponse
from .room_members import RoomMembersResponse
from .room_membership import UserJoinRoomResponse, UserJoinedRoomsResponse
from .room_messages import RoomMessagesResponse
from .server_capabilities import ServerCapabilitiesResponse
from .sync import SentEventResponse, SyncResponse
from .whoamiresponse import WhoAmIResponse



from typing import Optional

from aiomatrix import types
from ..syncing import SyncingAPI


def _add_reply(content: dict, reply_to: types.primitives.EventID):
    content['m.relates_to'] = {
        'm.in_reply_to': {
            'event_id': reply_to
        }
    }
    return content


class InstantMessagingAPI:
    def __init__(self, sync_api_client: SyncingAPI):
        self._sync_client = sync_api_client

    async def send_reaction(
            self, room_id: types.primitives.RoomID, event_id: types.primitives.EventID, key: str
    ) -> types.responses.SentEventResponse:
        content = {
            'm.relates_to': {
                'rel_type': 'm.annotation',
                'event_id': event_id,
                'key':      key
            }
        }
        r = await self._sync_client.send_message_event(room_id, types.misc.RoomEventTypesEnum.reaction, content)
        return r

    async def send_message(
            self, room_id: types.primitives.RoomID, text: str, reply_to: Optional[types.primitives.EventID] = None
    ) -> types.responses.SentEventResponse:
        content = {
            'body':    text,
            'msgtype': types.misc.RoomMessageEventMsgTypesEnum.text
        }
        if reply_to is not None:
            content = _add_reply(content, reply_to)
        r = await self._sync_client.send_message_event(room_id, types.misc.RoomEventTypesEnum.room_message, content)
        return r

    async def send_notice(
            self, room_id: types.primitives.RoomID, text: str, reply_to: Optional[types.primitives.EventID] = None
    ) -> types.responses.SentEventResponse:
        content = {
            'body':    text,
            'msgtype': types.misc.RoomMessageEventMsgTypesEnum.notice
        }
        if reply_to is not None:
            content = _add_reply(content, reply_to)
        r = await self._sync_client.send_message_event(room_id, types.misc.RoomEventTypesEnum.room_message, content)
        return r

    async def edit_text(
            self, room_id: types.primitives.RoomID, event_id: types.primitives.EventID, text: str
    ):
        content = {
            'body':          text,
            'msgtype':       types.misc.RoomMessageEventMsgTypesEnum.text,
            'm.new_content': {
                'body':    text,
                'msgtype': types.misc.RoomMessageEventMsgTypesEnum.text
            },
            'm.relates_to':  {
                'rel_type': 'm.replace',
                'event_id': event_id
            }
        }

        r = await self._sync_client.send_message_event(room_id, types.misc.RoomEventTypesEnum.room_message, content)
        return r



import datetime
from typing import Optional

import pydantic

from aiomatrix import types


class EventInDB(pydantic.BaseModel):
    account_id: types.primitives.UserID
    event_id: types.primitives.EventID
    relates_to: Optional[types.primitives.EventID]
    event_type: str
    room_id: Optional[types.primitives.RoomID]
    sender: types.primitives.UserID
    ts: datetime.datetime
    data: types.events.RoomEvent


class SeenEvent(pydantic.BaseModel):
    account_id: types.primitives.UserID
    event_id: types.primitives.EventID
